from MetaRagTool.RAG.Chunkers import ChunkingMethod, ChunkerFactory
import gradio as gr
from MetaRagTool.Utils.MRUtils import read_pdf, init_hf, listToString
import MetaRagTool.Utils.DataLoader as DataLoader
from MetaRagTool.RAG.MetaRAG import MetaRAG
import MetaRagTool.Constants as Constants
from MetaRagTool.LLM.GoogleGemini import Gemini

colors = [
    "#FF6B6B", "#4ECDC4", "#45B7D1", "#96CEB4", "#FFEEAD",
    "#D4A5A5", "#9B59B6", "#3498DB", "#E74C3C", "#2ECC71"
]
chunking_methods = [ChunkingMethod.SENTENCE_MERGER, ChunkingMethod.SENTENCE_MERGER_CROSS_PARAGRAPH,
                    ChunkingMethod.PARAGRAPH, ChunkingMethod.RECURSIVE, ChunkingMethod.SENTENCE]

rag: MetaRAG = None
encoder_model = None
reranker_model = None
contexts = None
qa = None

def load_models_and_data(encoder_name, reranker_name=None):
    from MetaRagTool.Encoders.SentenceTransformerEncoder import SentenceTransformerEncoder
    from MetaRagTool.Encoders.Reranker import CrossEncoderReranker
    global encoder_model, reranker_model, contexts, qa

    encoder_model = SentenceTransformerEncoder(encoder_name)
    if reranker_name is not None and len(reranker_name) > 0:
        reranker_model = CrossEncoderReranker(reranker_name)
    else:
        reranker_model = None

    contexts, qa = DataLoader.loadWikiFaQa(sample_size=10)

def tokenize_and_colorize(encoder_name, reranker_name, pdf_files, text, chunking_method, chunk_size, max_sentence_len, ignore_pfd_line_breaks, language_mode, gemini_api_key):
    global rag
    
    Constants.lang = 'fa' if language_mode == "Farsi" else 'multi'

    load_models_and_data(encoder_name, reranker_name)

    corpus_texts = []

    if pdf_files is not None:
        for pdf_file in pdf_files:
            corpus_texts.append(read_pdf(pdf_file.name, ignore_line_breaks=ignore_pfd_line_breaks))
    if text:
        corpus_texts.append(text)
    if not corpus_texts:
        corpus_texts.append(contexts[1])

    chunking_method = ChunkingMethod[chunking_method]

    api_key_to_use = gemini_api_key if gemini_api_key else Constants.API_KEY_GEMINI
    llm = Gemini(api_key=api_key_to_use)

    rag = MetaRAG(encoder_model=encoder_model, llm=llm, splitting_method=chunking_method,
                  chunk_size=chunk_size, max_sentence_len=max_sentence_len, reranker_model=reranker_model)

    rag.add_corpus(corpus_texts)
    tokens = rag.ChunksList

    color_index = 0
    colored_tokens = []
    for token in tokens:
        color = colors[color_index]
        color_index = (color_index + 1) % len(colors)
        colored_tokens.append((f"{token}", color))

    return rag.chunker.chunking_report(), colored_tokens

def run_chunker_only(pdf_files, text_input, chunking_method_name, chunk_size, max_sentence_len, ignore_pfd_line_breaks, language_mode):
    Constants.lang = 'fa' if language_mode == "Farsi" else 'multi'

    corpus_texts = []

    if pdf_files is not None:
        for pdf_file in pdf_files:
            corpus_texts.append(read_pdf(pdf_file.name, ignore_line_breaks=ignore_pfd_line_breaks))
    if text_input and len(text_input.strip()) > 0:
        corpus_texts.append(text_input)
    
    if not corpus_texts:
        contexts, qa = DataLoader.loadWikiFaQa(sample_size=10)
        corpus_texts.append(contexts[1])

    selected_chunking_method = ChunkingMethod[chunking_method_name]
    
    chunker_instance = ChunkerFactory.create_chunker(
        splitting_method=selected_chunking_method,chunksList=[],
        chunk_size=chunk_size)

    all_chunks = []
    all_chunks = chunker_instance.chunk_texts(corpus_texts)
        
    report_figure = chunker_instance.chunking_report()

    color_index = 0
    colored_chunks_display = []
    for chunk_text in all_chunks:
        color = colors[color_index % len(colors)]
        colored_chunks_display.append((f"{chunk_text}", color))
        color_index += 1
        
    return report_figure, colored_chunks_display


def retrieve_chunks(query, k, add_neighbor_chunks_smart, replace_retrieved_chunks_with_parent_paragraph, rerank):
    global rag

    if rag is None:
        return [("Please run the chunker first to initialize the RAG system.", "red")]

    try:
        rag.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        rag.replace_retrieved_chunks_with_parent_paragraph = replace_retrieved_chunks_with_parent_paragraph
        rag.rerank = rerank
        results = rag.retrieve(query, top_k=k)

        colored_chunks = []
        for i, chunk in enumerate(results):
            color = colors[i % len(colors)]
            colored_chunks.append((f"{chunk}\n", color))

        return colored_chunks

    except Exception as e:
        return [(f"Error during retrieval: {str(e)}", "red")]

def full_rag_ask(query, k, add_neighbor_chunks_smart, replace_retrieved_chunks_with_parent_paragraph, rerank):
    global rag

    if rag is None:
        return "Please run the chunker first to initialize the RAG system."

    try:
        rag.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        rag.replace_retrieved_chunks_with_parent_paragraph = replace_retrieved_chunks_with_parent_paragraph
        rag.rerank = rerank

        result = rag.ask(query, top_k=k)
        messages_history = listToString(rag.llm.messages_history, separator="\n\n")

        return result, messages_history

    except Exception as e:
        return f"Error during RAG processing: {str(e)}"

def full_tool_rag_ask(query, add_neighbor_chunks_smart, replace_retrieved_chunks_with_parent_paragraph, rerank):
    global rag

    if rag is None:
        return "Please run the chunker first to initialize the RAG system."

    try:
        rag.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        rag.replace_retrieved_chunks_with_parent_paragraph = replace_retrieved_chunks_with_parent_paragraph
        rag.rerank = rerank

        result = rag.ask(query, useTool=True)
        messages_history = listToString(rag.llm.messages_history, separator="\n\n")

        return result, messages_history

    except Exception as e:
        return f"Error during RAG processing: {str(e)}"

def load_app(encoder_model_name='sentence-transformers/LaBSE', reranker_model_name='cross-encoder/mmarco-mMiniLMv2-L12-H384-v1'):

    if gr.__version__ < '5.29.1':
        print("Please update gradio to version 5.29.1 or higher.")

    if Constants.HFToken is None:
        import os
        Constants.HFToken = os.getenv('HFToken')
        Constants.API_KEY_GEMINI = os.getenv('GEMINIToken')
        Constants.API_KEY_OPENROUTER = os.getenv('OPENROUTERToken')

    init_hf()



    chunker = gr.Interface(
        fn=tokenize_and_colorize,
        inputs=[
            gr.Textbox(
                label="Encoder Model Name",
                placeholder="Enter encoder model name (e.g., sentence-transformers/all-MiniLM-L6-v2)",
                value=encoder_model_name
            ),
            gr.Textbox(
                label="Reranker Model Name (Optional)",
                placeholder="Enter reranker model name (e.g., cross-encoder/ms-marco-MiniLM-L-6-v2)",
                value=reranker_model_name
            ),
            gr.File(
                label="Upload PDF", file_count="multiple"
            ),
            gr.Textbox(
                label="Or Enter your text",
                placeholder="Type some large amount of text here...",
                lines=3
            ),
            gr.Dropdown(
                label="Select Chunking Method",
                choices=[method.name for method in chunking_methods],
                value=chunking_methods[0].name
            ),
            gr.Slider(
                label="Select Chunk Size",
                minimum=1,
                maximum=300,
                step=1,
                value=90
            ),
            gr.Slider(
                label="Select Max Sentence Size",
                minimum=-1,
                maximum=500,
                step=1,
                value=-1
            ),
            gr.Checkbox(
                label="ignore_pfd_line_breaks",
                value=True
            ),
            gr.Radio(
                choices=["Farsi", "Multilingual"],
                label="Language Mode",
                value="Farsi"
            ),
            gr.Textbox(
                label="Gemini API Key (Optional)",
                placeholder="Enter your Gemini API Key here...",
                type="password"
            )
        ],
        outputs=[
            gr.Plot(label="Chunking Report"),
            gr.HighlightedText(
                label="Tokenized Output",
                show_inline_category=False,
                elem_classes=["rtl-text-display"],
                rtl=Constants.lang == 'fa'

            )
        ],
        title="Persian RAG",
        description="Enter some text and see it tokenized with different colors for each chunk!",
        theme="default",
    )

    chunkers_only_interface = gr.Interface(
        fn=run_chunker_only,
        inputs=[
            gr.File(
                label="Upload PDF", file_count="multiple"
            ),
            gr.Textbox(
                label="Or Enter your text",
                placeholder="Type some large amount of text here...",
                lines=3
            ),
            gr.Dropdown(
                label="Select Chunking Method",
                choices=[method.name for method in chunking_methods],
                value=chunking_methods[0].name
            ),
            gr.Slider(
                label="Select Chunk Size",
                minimum=1,
                maximum=300,
                step=1,
                value=90
            ),
            gr.Slider(
                label="Select Max Sentence Size",
                minimum=-1,
                maximum=500,
                step=1,
                value=-1
            ),
            gr.Checkbox(
                label="Ignore PDF line breaks",
                value=True
            ),
            gr.Radio(
                choices=["Farsi", "Multilingual"],
                label="Language Mode",
                value="Farsi"
            )
        ],
        outputs=[
            gr.Plot(label="Chunking Report"),
            gr.HighlightedText(
                label="Chunked Output",
                show_inline_category=False,
                elem_classes=["rtl-text-display"],
                rtl=Constants.lang == 'fa'

            )
        ],
        title="Chunkers Only",
        description="Select a chunking method, input text/PDF, and see the chunked output directly. Note: Some methods might have limitations with nlp_models=[].",
        theme="default",
    )

    retriever = gr.Interface(
        fn=retrieve_chunks,
        inputs=[
            gr.Textbox(
                label="Enter your query",
                placeholder="Type some text here...",
                lines=3
            ),
            gr.Slider(
                label="Select K",
                minimum=1,
                maximum=100,
                step=1,
                value=10
            ),
            gr.Checkbox(
                label="Include Neighbors",
                value=False
            ),
            gr.Checkbox(
                label="Replace With Parent Paragraph",
                value=False
            ),
            gr.Checkbox(
                label="Rerank",
                value=False
            )
        ],
        outputs=[
            gr.HighlightedText(
                label="retrieved chunks",
                show_inline_category=False,
                elem_classes=["rtl-text-display"],
                rtl=Constants.lang == 'fa'

            )
        ],
        title="Retriever with Colored Output",
        theme="default",
    )

    full_rag = gr.Interface(
        fn=full_rag_ask,
        inputs=[
            gr.Textbox(
                label="Enter your query",
                placeholder="Type some text here...",
                lines=3
            ),
            gr.Slider(
                label="Select K",
                minimum=1,
                maximum=100,
                step=1,
                value=10
            ),
            gr.Checkbox(
                label="Include Neighbors",
                value=False
            ),
            gr.Checkbox(
                label="Replace With Parent Paragraph",
                value=False
            ),
            gr.Checkbox(
                label="Rerank",
                value=False
            )
        ],
        outputs=[
            gr.Textbox(
                label="RAG Output",
                lines=20,
                elem_classes=["rtl-text-display"],
                rtl=Constants.lang == 'fa'

            ),
            gr.Textbox(
                label="LLM Messages History",
                lines=20,
                elem_classes=["rtl-text-display"]
            )
        ],
        title="Full RAG with Raw Output",
        theme="default",
    )

    full_tool_rag = gr.Interface(
        fn=full_tool_rag_ask,
        inputs=[
            gr.Textbox(
                label="Enter your query",
                placeholder="Type some text here...",
                lines=3
            ),
            gr.Checkbox(
                label="Include Neighbors",
                value=False
            ),
            gr.Checkbox(
                label="Replace With Parent Paragraph",
                value=False
            ),
            gr.Checkbox(
                label="Rerank",
                value=False
            )
        ],
        outputs=[
            gr.Textbox(
                label="RAG Output",
                lines=20,
                elem_classes=["rtl-text-display"],
                rtl=Constants.lang == 'fa'
            ),
            gr.Textbox(
                label="LLM Messages History",
                lines=20,
                elem_classes=["rtl-text-display"]
            )
        ],
        title="Full Tool RAG with Raw Output",
        theme="default",
    )

    iface = gr.TabbedInterface(
        [chunker,  retriever, full_rag, full_tool_rag,chunkers_only_interface],
        ["Chunk and Encode", "Retriever", "Full RAG", "Full Tool RAG","Chunk Only"]
    )

    iface.launch(show_error=True)









class MRDocument:
    def __init__(self,text=""):
        self.Paragraphs = []
        self.Sentences = []
        self.Text = text
        self.isChunked = False

    def AddParagraph(self, paragraph):
        if paragraph is None:
            print( "Warning: paragraph is None")
            return
        self.Paragraphs.append(paragraph)

    def AddSentence(self, sentence):
        if sentence is None:
            print( "Warning: sentence is None")
            return
        self.Sentences.append(sentence)



class MRParagraph:
    def __init__(self,document,text=""):
        self.Document = document
        self.Sentences = []
        self.Prev = None
        self.Next = None
        self.Text = text
        self.Embeddings = None

    def AddSentence(self, sentence):
        if sentence is None:
            print( "Warning: sentence is None")
            return
        self.Sentences.append(sentence)

    def SetPrev(self, prev):
        self.Prev = prev

    def SetNext(self, nextParagraph):
        self.Next = nextParagraph


class MRSentence:
    def __init__(self ,document,paragraph,text=""):
        self.Document = document
        self.Paragraph = paragraph
        self.Prev = None
        self.Next = None
        self.Text = text

    def SetPrev(self, prev):
        self.Prev = prev

    def SetNext(self, nextSentence):
        self.Next = nextSentence

class MRChunk:
    def __init__(self,document, text=""):
        self.Text = text
        self.Paragraphs = []
        self.Sentences = []
        self.Document = document
        self.PrevRelated = None
        self.NextRelated = None
        self.Embeddings = None
        self.Length = -1

    def AddParagraph(self, paragraph):
        if paragraph is None:
            print( "Warning: paragraph is None")
            return
        if paragraph not in self.Paragraphs:
            self.Paragraphs.append(paragraph)

    def AddSentence(self, sentence):
        if sentence is None:
            print( "Warning: sentence is None")
            return
        self.Sentences.append(sentence)



    def __str__(self):
        return self.Text

    def __repr__(self):
        return self.Text


def generate_document_structure(rawDocumentsTexts:list, DocumentsList:list, max_sentence_len:int):
    from MetaRagTool.Utils.MRUtils import capped_sent_tokenize

    if max_sentence_len == -1:
        print("Error: max_sentence_len is set to -1, which is invalid.")


    # Generate the document structure
    for documentText in rawDocumentsTexts:
        if len(documentText) < 1:
            continue

        newDoc = MRDocument(documentText)
        DocumentsList.append(newDoc)

        prevParagraph = None
        for paragraphText in documentText.split('\n'):
            if len(paragraphText) < 5:
                continue

            newParagraph = MRParagraph(document=newDoc, text=paragraphText)

            prevSentence = None
            sentences = capped_sent_tokenize(paragraphText, max_sentence_len)
            if len(sentences) < 1:
                continue

            for sentenceText in sentences:

                newSentence = MRSentence(document=newDoc, paragraph=newParagraph, text=sentenceText)
                newParagraph.AddSentence(newSentence)
                newDoc.AddSentence(newSentence)
                if prevSentence is not None:
                    prevSentence.SetNext(newSentence)
                    newSentence.SetPrev(prevSentence)

                prevSentence = newSentence

            newDoc.AddParagraph(newParagraph)
            if prevParagraph is not None:
                prevParagraph.SetNext(newParagraph)
                newParagraph.SetPrev(prevParagraph)

            prevParagraph = newParagraph

    allParagraphs = [paragraph for doc in DocumentsList for paragraph in doc.Paragraphs]
    allSentences = [sentence for doc in DocumentsList for sentence in doc.Sentences]

    return allParagraphs, allSentences



from abc import ABC, abstractmethod
from enum import Enum
from MetaRagTool.RAG.DocumentStructs import MRDocument, MRChunk, MRSentence

class ChunkingMethod(Enum):
    SENTENCE_MERGER_CROSS_PARAGRAPH = 'sentence_merger_cross_paragraph'
    SENTENCE = 'sent'
    PARAGRAPH = 'paragraph'
    DOCUMENT = 'document'
    RECURSIVE = 'recursive'
    SENTENCE_MERGER = 'sentence_merger'

class Chunker(ABC):
    def __init__(self, splitting_method: ChunkingMethod, chunksList:list, chunk_size: int = 90, chunk_overlap: int = 3):
        self.splitting_method = splitting_method
        self.chunksList = chunksList
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap


    def chunk_texts(self, texts: list):
        from MetaRagTool.RAG.DocumentStructs import generate_document_structure
        documentLists=[]
        allParagraphs,allSentences=generate_document_structure(rawDocumentsTexts=texts, DocumentsList=documentLists, max_sentence_len=self.chunk_size)
        print(f"Corpus structure: {len(documentLists)} documents and {len(allParagraphs)} paragraphs and {len(allSentences)} sentences")

        return self.chunk_documents(documentLists)



    def chunk_documents(self, documents: list) -> list:
        for doc in documents:
            if doc.isChunked:
                continue

            doc.isChunked = True
            self._chunk_document_core(doc)


        for chunk in self.chunksList:
            chunk.Length = token_len(chunk.Text)

        return self.chunksList



    @abstractmethod
    def _chunk_document_core(self, document: MRDocument):
        pass

    def chunking_report(self):
        import matplotlib.pyplot as plt
        import numpy as np

        chunk_sizes = [token_len(chunk.Text) for chunk in self.chunksList]
        total_chunks = len(chunk_sizes)
        avg_chunk_size = sum(chunk_sizes) / total_chunks if total_chunks > 0 else 0
        max_chunk_size = max(chunk_sizes) if total_chunks > 0 else 0
        min_chunk_size = min(chunk_sizes) if total_chunks > 0 else 0
        median_chunk_size = np.median(chunk_sizes) if total_chunks > 0 else 0
        std_chunk_size = np.std(chunk_sizes) if total_chunks > 0 else 0
        # chunk_size_distribution = np.histogram(chunk_sizes, bins='auto') if total_chunks > 0 else ([], [])

        print(f"Chunking Report:")
        print(f"Total number of chunks: {total_chunks}")
        print(f"Average chunk size: {avg_chunk_size:.2f} tokens")
        print(f"Maximum chunk size: {max_chunk_size} tokens")
        print(f"Minimum chunk size: {min_chunk_size} tokens")
        print(f"Median chunk size: {median_chunk_size:.2f} tokens")
        print(f"Standard deviation of chunk size: {std_chunk_size:.2f} tokens")

        plt.clf()
        plt.hist(chunk_sizes, bins='auto', edgecolor='black')
        plt.title('Chunk Size Distribution')
        plt.xlabel('Chunk Size (tokens)')
        plt.ylabel('Frequency')
        plt.show()

        return plt





class SentenceChunker(Chunker):
    def __init__(self, chunksList:list, chunk_size: int = 90, chunk_overlap: int = 3):
        super().__init__(ChunkingMethod.SENTENCE, chunksList, chunk_size, chunk_overlap)


    def _chunk_document_core(self, document: MRDocument):
        prevChunk = None
        for sentence in document.Sentences:
            newChunk = MRChunk(document=document, text=sentence.Text)
            newChunk.AddParagraph(sentence.Paragraph)
            newChunk.AddSentence(sentence)
            self.chunksList.append(newChunk)
            if prevChunk is not None:
                prevChunk.NextRelated = newChunk
                newChunk.PrevRelated = prevChunk
            prevChunk = newChunk

class ParagraphChunker(Chunker):
    def __init__(self, chunksList:list, chunk_size: int = 90, chunk_overlap: int = 3):
        super().__init__(ChunkingMethod.PARAGRAPH, chunksList, chunk_size, chunk_overlap)

    def _chunk_document_core(self, document: MRDocument):
        prevChunk = None
        for paragraph in document.Paragraphs:
            newChunk = MRChunk(document=document, text=paragraph.Text)
            newChunk.AddParagraph(paragraph)
            for sentence in paragraph.Sentences:
                newChunk.AddSentence(sentence)

            self.chunksList.append(newChunk)
            if prevChunk is not None:
                prevChunk.NextRelated = newChunk
                newChunk.PrevRelated = prevChunk
            prevChunk = newChunk


class DocumentChunker(Chunker):
    def __init__(self, chunksList:list, chunk_size: int = 90, chunk_overlap: int = 3):
        super().__init__(ChunkingMethod.DOCUMENT, chunksList, chunk_size, chunk_overlap)

    def _chunk_document_core(self, document: MRDocument):
        newChunk = MRChunk(document=document, text=document.Text)
        for paragraph in document.Paragraphs:
            newChunk.AddParagraph(paragraph)
            for sentence in paragraph.Sentences:
                newChunk.AddSentence(sentence)

        self.chunksList.append(newChunk)



class SentenceMergerChunker(Chunker):
    def __init__(self, chunksList:list, chunk_size: int = 90, chunk_overlap: int = 3, sent_merge_merged_chunks: bool = True):
        super().__init__(ChunkingMethod.SENTENCE_MERGER, chunksList, chunk_size, chunk_overlap)
        self.sent_merge_merged_chunks = sent_merge_merged_chunks

    def _sent_merger_create_chunk(self, currentChunkText, current_chunk_sentences, document, paragraph, prevChunk):
        if self.sent_merge_merged_chunks and prevChunk is not None and token_len(prevChunk.Text + currentChunkText) < self.chunk_size:
            prevChunk.Text += currentChunkText
            for s in current_chunk_sentences:
                prevChunk.AddSentence(s)
            prevChunk.AddParagraph(paragraph)
        else:
            newChunk = MRChunk(document=document, text=currentChunkText)
            newChunk.AddParagraph(paragraph)
            for s in current_chunk_sentences:
                newChunk.AddSentence(s)
            self.chunksList.append(newChunk)
            if prevChunk is not None:
                prevChunk.NextRelated = newChunk
                newChunk.PrevRelated = prevChunk
            prevChunk = newChunk
        currentChunkText = ""
        current_chunk_sentences = []



        return currentChunkText, current_chunk_sentences, prevChunk




    def _chunk_document_core(self, document: MRDocument):
        prevChunk = None

        for paragraph in document.Paragraphs:
            current_chunk_sentences = []
            currentChunkText = ""
            currentSentence: MRSentence = paragraph.Sentences[0]
            while currentSentence is not None:
                nextText = currentSentence.Text + " "
                # If adding this sentence would exceed chunk_size, finalize current chunk first.
                if token_len(currentChunkText + nextText) > self.chunk_size and currentChunkText.strip():
                    currentChunkText, current_chunk_sentences, prevChunk = self._sent_merger_create_chunk(
                        currentChunkText,
                        current_chunk_sentences,
                        document, paragraph,
                        prevChunk)

                currentChunkText += nextText
                current_chunk_sentences.append(currentSentence)

                # If next sentence is None or we've just reached chunk size, finalize the chunk.
                if currentSentence.Next is None:
                    currentChunkText, current_chunk_sentences, prevChunk = self._sent_merger_create_chunk(
                        currentChunkText,
                        current_chunk_sentences,
                        document, paragraph,
                        prevChunk)

                currentSentence = currentSentence.Next



class SentenceMergerCrossParagraphChunker(Chunker):
    def __init__(self, chunksList:list, chunk_size: int = 90, chunk_overlap: int = 3):
        super().__init__(ChunkingMethod.SENTENCE_MERGER_CROSS_PARAGRAPH, chunksList, chunk_size, chunk_overlap)

    def _chunk_document_core(self, document: MRDocument):
        prevChunk = None
        current_chunk_sentences = []
        currentChunkText = ""

        for idx, sentt in enumerate(document.Sentences):
            # If adding this sentence exceeds chunk_size, create the chunk right away
            if currentChunkText and token_len(currentChunkText + sentt.Text) > self.chunk_size:
                newChunk = MRChunk(document=document, text=currentChunkText)
                for s in current_chunk_sentences:
                    newChunk.AddSentence(s)
                    newChunk.AddParagraph(s.Paragraph)
                self.chunksList.append(newChunk)
                if prevChunk is not None:
                    prevChunk.NextRelated = newChunk
                    newChunk.PrevRelated = prevChunk

                prevChunk = newChunk
                currentChunkText = ""
                current_chunk_sentences = []

            currentChunkText += sentt.Text + " "
            current_chunk_sentences.append(sentt)

            # If it's the last sentence, finalize the current chunk
            if idx == len(document.Sentences) - 1:
                newChunk = MRChunk(document=document, text=currentChunkText)
                for s in current_chunk_sentences:
                    newChunk.AddSentence(s)
                    newChunk.AddParagraph(s.Paragraph)
                self.chunksList.append(newChunk)
                if prevChunk is not None:
                    prevChunk.NextRelated = newChunk
                    newChunk.PrevRelated = prevChunk


class RecursiveChunker(Chunker):
    def __init__(self, chunksList:list, chunk_size: int = 90, chunk_overlap: int = 3):
        from langchain_text_splitters import RecursiveCharacterTextSplitter

        super().__init__(ChunkingMethod.RECURSIVE, chunksList, chunk_size, chunk_overlap)
        self.splitter = RecursiveCharacterTextSplitter(
            chunk_size=self.chunk_size,
            chunk_overlap=self.chunk_overlap,
            length_function=token_len,
            is_separator_regex=False,
        )

    def _chunk_document_core(self, document: MRDocument):
        text_chunks__for_this_doc = self.splitter.split_text(document.Text)
        for text_chunk in text_chunks__for_this_doc:
            self.chunksList.append(MRChunk(document=document, text=text_chunk))


class ChunkerFactory:
    @staticmethod
    def create_chunker(splitting_method: ChunkingMethod, chunksList: list, chunk_size: int = 90, chunk_overlap: int = 3,sent_merge_merged_chunks: bool = True) -> Chunker:
        if splitting_method == ChunkingMethod.SENTENCE:
            return SentenceChunker(chunksList, chunk_size, chunk_overlap)
        elif splitting_method == ChunkingMethod.PARAGRAPH:
            return ParagraphChunker(chunksList, chunk_size, chunk_overlap)
        elif splitting_method == ChunkingMethod.DOCUMENT:
            return DocumentChunker(chunksList, chunk_size, chunk_overlap)
        elif splitting_method == ChunkingMethod.SENTENCE_MERGER:
            return SentenceMergerChunker(chunksList, chunk_size, chunk_overlap, sent_merge_merged_chunks)
        elif splitting_method == ChunkingMethod.SENTENCE_MERGER_CROSS_PARAGRAPH:
            return SentenceMergerCrossParagraphChunker(chunksList, chunk_size, chunk_overlap)
        elif splitting_method == ChunkingMethod.RECURSIVE:
            return RecursiveChunker(chunksList, chunk_size, chunk_overlap)
        else:
            raise ValueError(f"Unsupported chunking method: {splitting_method}")







from MetaRagTool.Utils.MRUtils import token_len



import MetaRagTool.Constants
import MetaRagTool.LLM
import MetaRagTool.Encoders
import MetaRagTool.RAG
import MetaRagTool.Utils
import MetaRagTool.Evaluations
import MetaRagTool.Apps


from MetaRagTool.RAG.MetaRAG import MetaRAG
from MetaRagTool.LLM.GoogleGemini import Gemini
from MetaRagTool.LLM.OpenaiGpt import OpenaiGpt
from MetaRagTool.Encoders.SentenceTransformerEncoder import SentenceTransformerEncoder
from MetaRagTool.Evaluations import Evals













__version__ = "0.3.11"
print("MetaRagTool imported")



from MetaRagTool.Encoders.Encoder import Encoder,OneByOneEncoder
from transformers import XLMRobertaModel
from transformers import AutoConfig, AutoTokenizer, AutoModel

class AutoModelEncoder(Encoder):

    def __init__(self, model_name: str, verbose=False):
        super().__init__(model_name, verbose)
        self.model = AutoModel.from_pretrained(model_name, trust_remote_code=True)
        print("Model loaded successfully")

    def encode(self, sentences, isQuery=True):
        embeddings = self.model.encode(sentences,
                                       show_progress_bar=not isQuery,
                                       )

        return embeddings


class XLMRobertaEncoder(OneByOneEncoder):
    def __init__(self, model_name: str, verbose=False):
        super().__init__(model_name, verbose)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = XLMRobertaModel.from_pretrained(model_name, add_pooling_layer=False).to('cuda')

        print("Model loaded successfully")

    def _encode_str(self, sentences: str):
        tokenized_input = self.tokenizer(sentences, return_tensors='pt').to('cuda')
        outputs = self.model(**tokenized_input)
        sentence_representation = outputs.last_hidden_state[:, 0]
        sentence_representation = sentence_representation.detach().cpu().numpy()
        return sentence_representation


class BertEncoder(OneByOneEncoder):

    def __init__(self, model_name: str, verbose=False):
        super().__init__(model_name, verbose)
        self.config = AutoConfig.from_pretrained(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModel.from_pretrained(model_name).to('cuda')
        print("Model loaded successfully")

    def _encode_str(self, sentences: str):
        tokens = self.tokenizer(sentences, return_tensors='pt').to('cuda')
        output = self.model(**tokens)
        pooler_output = output.pooler_output
        return pooler_output.detach().cpu().numpy()



import time
import faiss
import numpy as np
from hazm import Normalizer

from MetaRagTool.RAG.Chunkers import ChunkerFactory
from MetaRagTool.RAG.DocumentStructs import MRSentence, MRDocument, MRParagraph, generate_document_structure
from MetaRagTool.RAG.Chunkers import ChunkingMethod

from MetaRagTool.Utils.MRUtils import capped_sent_tokenize, token_len, reflect_vector, remove_duplicates

from MetaRagTool.LLM.LLMIdentity import LLMIdentity
from MetaRagTool.Encoders.Encoder import Encoder
from MetaRagTool.Encoders.Reranker import Reranker



class MetaRAG:


    def __init__(self, encoder_model: Encoder, llm: LLMIdentity=None, reranker_model: Reranker = None,
                 splitting_method=ChunkingMethod.SENTENCE_MERGER, chunk_size=90, chunk_overlap=3, max_sentence_len=-1,
                 use_neighbor_embeddings=False
                 , use_parentParagraph_embeddings=False, add_neighbor_chunks=False, add_neighbor_chunks_smart=False,
                 breath_first_retrival=False
                 , depth_first_retrival=False,
                 replace_retrieved_chunks_with_parent_paragraph=False, normalize_text=False, normalizer=Normalizer(),
                 rerank=False,
                 additional_top_k=5, include_reflections=False,
                 include_refractions=False, sent_merge_merged_chunks=True, log_chunking_report=True, weighted_bfs=False,
                 embedding_steering_influence_factor=0.35, add_neighbor_chunks_k=2):



        self.add_neighbor_chunks_k = add_neighbor_chunks_k
        self.log_chunking_report = log_chunking_report
        self.additional_top_k = additional_top_k
        if max_sentence_len == -1:
            max_sentence_len = chunk_size
        self.max_sentence_len = max_sentence_len
        self.llm = llm
        self.splitting_method = splitting_method
        self.chunk_size = chunk_size
        self.encoder_model = encoder_model
        self.chunk_overlap = chunk_overlap
        self.index = None
        self.DocumentsList = []
        self.ChunksList = []
        self.use_neighbor_embeddings = use_neighbor_embeddings
        self.use_parentParagraph_embeddings = use_parentParagraph_embeddings
        self.add_neighbor_chunks = add_neighbor_chunks
        self.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        self.breath_first_retrival = breath_first_retrival
        self.depth_first_retrival = depth_first_retrival
        self.replace_retrieved_chunks_with_parent_paragraph = replace_retrieved_chunks_with_parent_paragraph
        self.normalize_text = normalize_text
        self.normalizer = normalizer
        self.rerank = rerank
        self.reranker_model = reranker_model
        self.include_reflections = include_reflections
        self.include_refractions = include_refractions
        self.weighted_bfs = weighted_bfs
        self.embedding_steering_influence_factor=embedding_steering_influence_factor

        # merge small chunks into normal ones
        self.sent_merge_merged_chunks = sent_merge_merged_chunks
        self.time_to_encode_corpus = 0
        self.answer_mode='none'

        self.chunker = None





    def report(self):
        default_values = {
            'splitting_method': None,
            'chunk_size': None,
            'chunk_overlap': 3,
            'max_sentence_len': 90,
            'use_neighbor_embeddings': False,
            'use_parentParagraph_embeddings': False,
            'add_neighbor_chunks': False,
            'add_neighbor_chunks_smart': False,
            'breath_first_retrival': False,
            'depth_first_retrival': False,
            'replace_retrieved_chunks_with_parent_paragraph': False,
            'normalize_text': False,
            'rerank': False,
            'additional_top_k': 5,
            'include_reflections': False,
            'include_refractions': False,
            'sent_merge_merged_chunks': True,
        }

        for param, default_value in default_values.items():
            current_value = getattr(self, param)
            if current_value != default_value:
                print(f"{param}: {current_value}")

    def apply_config(self, ragConfig):

        self.splitting_method = ragConfig.splitting_method
        self.chunk_size = ragConfig.chunk_size
        self.chunk_overlap = ragConfig.chunk_overlap
        self.max_sentence_len = ragConfig.max_sentence_len
        if self.max_sentence_len == -1:
            self.max_sentence_len = self.chunk_size
        self.use_neighbor_embeddings = ragConfig.use_neighbor_embeddings
        self.use_parentParagraph_embeddings = ragConfig.use_parentParagraph_embeddings
        self.add_neighbor_chunks = ragConfig.add_neighbor_chunks
        self.add_neighbor_chunks_smart = ragConfig.add_neighbor_chunks_smart
        self.breath_first_retrival = ragConfig.breath_first_retrival
        self.depth_first_retrival = ragConfig.depth_first_retrival
        self.replace_retrieved_chunks_with_parent_paragraph = ragConfig.replace_retrieved_chunks_with_parent_paragraph
        self.normalize_text = ragConfig.normalize_text
        self.include_reflections = ragConfig.include_reflections
        self.include_refractions = ragConfig.include_refractions
        self.sent_merge_merged_chunks = ragConfig.sent_merge_merged_chunks
        self.log_chunking_report = ragConfig.log_chunking_report
        self.add_neighbor_chunks_k = ragConfig.add_neighbor_chunks_k
        self.embedding_steering_influence_factor = ragConfig.embedding_steering_influence_factor
        self.weighted_bfs = ragConfig.weighted_bfs
        self.additional_top_k = ragConfig.additional_top_k
        self.include_reflections_k = ragConfig.include_reflections_k
        self.rerank = ragConfig.rerank
        self.reranker_model = ragConfig.reranker

    def _encode_and_index_dataset(self):
        if self.use_parentParagraph_embeddings:
            self._set_paragraph_embeddings()

        chunks_to_encode = [chunk for chunk in self.ChunksList if chunk.Embeddings is None]
        print(f"Number of chunks to encode: {len(chunks_to_encode)}")
        chunks_texts = [self.normalizer.normalize(chunk.Text) if self.normalize_text else chunk.Text for chunk in
                        chunks_to_encode]


        t1= time.time()

        # Step 2: Encode documents using a pre-trained model
        document_embeddings = self.encoder_model.encode(chunks_texts, isQuery=False)
        # print("Chunk encoding completed")

        # Convert embeddings to numpy array
        document_embeddings = np.array(document_embeddings)
        document_embeddings_copy = np.array(document_embeddings)
        # for each chunk, add the embedding of the next and previous chunk to it with a weight of 0.5
        if self.use_neighbor_embeddings:
            for i, chunk in enumerate(chunks_to_encode):
                if chunk.PrevRelated is not None:
                    document_embeddings[i] = document_embeddings[i] + self.embedding_steering_influence_factor * document_embeddings_copy[i - 1]
                if chunk.NextRelated is not None:
                    document_embeddings[i] = document_embeddings[i] + self.embedding_steering_influence_factor * document_embeddings_copy[i + 1]

        if self.use_parentParagraph_embeddings:
            for i, chunk in enumerate(chunks_to_encode):
                parentParagraphEmbedding = np.mean([p.Embeddings for p in chunk.Paragraphs],axis=0)
                parentParagraphEmbedding = parentParagraphEmbedding / np.linalg.norm(parentParagraphEmbedding)
                document_embeddings[i] = document_embeddings[i] + self.embedding_steering_influence_factor * parentParagraphEmbedding

        for i, chunk in enumerate(chunks_to_encode):
            chunk.Embeddings = document_embeddings[i]
        # Step 3: Index embeddings using FAISS
        embedding_dimension = document_embeddings.shape[1]
        if self.index is None:
            self.index = faiss.IndexFlatL2(embedding_dimension)

        faiss.normalize_L2(document_embeddings)
        self.index.add(document_embeddings)  # Add document embeddings to the index

        self.time_to_encode_corpus += (time.time()-t1)
        # print("Chunk indexing completed")

    def _set_paragraph_embeddings(self):
        paragraphs_to_encode = [paragraph for paragraph in self.allParagraphs if paragraph.Embeddings is None]
        print(f"Number of paragraphs to encode: {len(paragraphs_to_encode)}")
        paragraphs_texts = [paragraph.Text for paragraph in paragraphs_to_encode]

        # Step 2: Encode paragraphs using a pre-trained model
        paragraph_embeddings = self.encoder_model.encode(paragraphs_texts, isQuery=False)
        print("Paragraph encoding completed")

        # Convert embeddings to numpy array
        paragraph_embeddings = np.array(paragraph_embeddings)

        for i, paragraph in enumerate(paragraphs_to_encode):
            paragraph.Embeddings = paragraph_embeddings[i]


    def _chunkify(self):

        if self.chunker is None:
            self.chunker = ChunkerFactory.create_chunker(splitting_method=self.splitting_method, chunksList=self.ChunksList,
                                                         chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap, sent_merge_merged_chunks=self.sent_merge_merged_chunks)

        self.chunker.chunk_documents(self.DocumentsList)

        if self.log_chunking_report:self.chunker.chunking_report()




    def add_corpus(self, raw_documents_text: list,encode=True):
        self.allParagraphs,self.allSentences=generate_document_structure(rawDocumentsTexts=raw_documents_text,DocumentsList=self.DocumentsList,
                                    max_sentence_len=self.max_sentence_len if self.max_sentence_len != -1 else self.chunk_size)

        print(f"Corpus structure: {len(self.DocumentsList)} documents and {len(self.allParagraphs)} paragraphs and {len(self.allSentences)} sentences")


        self._chunkify()
        if encode:self._encode_and_index_dataset()

    def _retrieve_core(self, query_embedding, top_k):
        if top_k < 1:
            print("Error: top_k is set to 0, which is invalid.")
            return []

        query_embedding = np.array(query_embedding).reshape(1, -1)
        distances, indices = self.index.search(query_embedding, top_k)
        retrieved_chunks = [self.ChunksList[i] for i in indices[0]]
        return retrieved_chunks


    def _check_feature_compatibility(self):
        if self.splitting_method == ChunkingMethod.RECURSIVE and (self.replace_retrieved_chunks_with_parent_paragraph or self.add_neighbor_chunks or self.add_neighbor_chunks_smart):
            print("Error: Recursive splitting method is not compatible with replace_retrieved_chunks_with_parent_paragraph. Please set it to False.")
            self.replace_retrieved_chunks_with_parent_paragraph = False
            self.add_neighbor_chunks_smart = False
            self.add_neighbor_chunks = False

    def retrieve(self, query, top_k=20, force_basics=False):
        # if sum_of_retrieved_token_length_limit>0:
        #     # fixing top_k
        #     top_k = sum_of_retrieved_token_length_limit // self.chunk_size
        #     if self.replace_retrieved_chunks_with_parent_paragraph:
        #         top_k = top_k // 2
        #     if self.include_reflections or self.include_refractions:
        #         top_k = top_k // 2
        #     if self.add_neighbor_chunks or self.add_neighbor_chunks_smart:
        #         top_k = top_k // 2
        #     if self.breath_first_retrival or self.depth_first_retrival or self.weighted_bfs:
        #         top_k = top_k // 2

        self._check_feature_compatibility()


        if top_k<1:
            print("Error: top_k is set to 0, which is invalid.")
            top_k=1

        # Encode the query
        if self.normalize_text:
            query = self.normalizer.normalize(query)

        query_embedding = self.encoder_model.encode([query])

        retrieved_chunks = self._retrieve_core(query_embedding, top_k*5 if self.rerank else top_k)


        if self.rerank:
            if self.reranker_model is None: print("Error: Reranker model is not set. Please set the Reranker model before using it or set rerank=False")

            retrieved_chunks=self.reranker_model.apply_rerank_MRChunks(query=query, chunks=retrieved_chunks)
            retrieved_chunks = retrieved_chunks[: top_k]


        if self.breath_first_retrival and not force_basics:
            additional_chunks = []
            top_k_retrieved_chunks = retrieved_chunks[:top_k // 5]
            for chunk in top_k_retrieved_chunks:
                additional_chunks.extend(self._retrieve_core(chunk.Embeddings, self.additional_top_k))
            retrieved_chunks.extend(additional_chunks)
            retrieved_chunks = remove_duplicates(retrieved_chunks)


        elif self.depth_first_retrival and not force_basics and top_k > 1:
            # repeat retrieval for the first chunk with the original top_k value, depth_first_retrival_k times, each time with the retrieved chunks (like going deeper only in one branch of the tree)
            additional_chunks = []
            # Replace the empty section with:
            current_chunk = retrieved_chunks[0]  # Start with first retrieved chunk
            for _ in range(2):
                level_chunks = self._retrieve_core(current_chunk.Embeddings, top_k // 2)
                additional_chunks.extend(level_chunks)
                current_chunk = level_chunks[0]  # Go deeper with first chunk

            retrieved_chunks.extend(additional_chunks)
            retrieved_chunks = remove_duplicates(retrieved_chunks)

        elif self.weighted_bfs and not force_basics:
            additional_chunks = []
            for i, chunk in enumerate(retrieved_chunks):
                chunk_top_k = top_k // (2 ** (i + 1))
                if chunk_top_k < 1:
                    break
                additional_chunks.extend(self._retrieve_core(chunk.Embeddings, chunk_top_k))

            retrieved_chunks.extend(additional_chunks)
            retrieved_chunks = remove_duplicates(retrieved_chunks)


        if self.include_reflections:
            # for each retrieved chunk, get its reflection based on query and retrived only 1 more chunks with the new vector
            additional_chunks = []
            for chunk in retrieved_chunks:
                reflection = reflect_vector(query_embedding[0], chunk.Embeddings)
                reflection = np.array([reflection], dtype=np.float32)
                faiss.normalize_L2(reflection)
                additional_chunks.extend(self._retrieve_core(reflection, 1))

            retrieved_chunks.extend(additional_chunks)
            retrieved_chunks = remove_duplicates(retrieved_chunks)

        elif self.include_refractions:
            # for each retrieved chunk, newV=q-v and retrive the closest chunk to newV
            additional_chunks = []
            for chunk in retrieved_chunks:
                newV = query_embedding[0] - chunk.Embeddings
                newV = np.array([newV], dtype=np.float32)
                faiss.normalize_L2(newV)

                closest_chunk = self._retrieve_core(newV, 1)
                additional_chunks.extend(closest_chunk)

            retrieved_chunks.extend(additional_chunks)
            retrieved_chunks = remove_duplicates(retrieved_chunks)

        if self.add_neighbor_chunks and not force_basics:
            additional_chunks = []
            for chunk in retrieved_chunks:
                if chunk.PrevRelated is not None:
                    additional_chunks.append(chunk.PrevRelated)
                if chunk.NextRelated is not None:
                    additional_chunks.append(chunk.NextRelated)
            retrieved_chunks.extend(additional_chunks)

        elif self.add_neighbor_chunks_smart and not force_basics:
            # only add the next and previous chunk if their embedding added to the original chunk embedding gets it closer(cosine sim) to query embedding
            additional_chunks = [chunk for chunk in retrieved_chunks]

            for chunk in retrieved_chunks:

                def check_add(neighbour_chunk,chunk0,current_depth):
                    current_depth+=1
                    orig_similarity = cosine_similarity(query_embedding[0], chunk0.Embeddings)
                    combined_embedding = (chunk0.Embeddings + neighbour_chunk.Embeddings) / 2
                    combined_similarity = cosine_similarity(query_embedding[0], combined_embedding)
                    if combined_similarity > orig_similarity:
                        if neighbour_chunk not in additional_chunks:
                            additional_chunks.append(neighbour_chunk)
                        if current_depth<self.add_neighbor_chunks_k:
                            if neighbour_chunk.PrevRelated is not None:
                                check_add(neighbour_chunk.PrevRelated,chunk0,current_depth)
                            if neighbour_chunk.NextRelated is not None:
                                check_add(neighbour_chunk.NextRelated,chunk0,current_depth)




                if chunk.PrevRelated is not None:
                    check_add(chunk.PrevRelated,chunk,0)

                if chunk.NextRelated is not None:
                    check_add(chunk.NextRelated,chunk,0)

            # remove retrieved_chunks from additional_chunks
            additional_chunks = [chunk for chunk in additional_chunks if chunk not in retrieved_chunks]


            retrieved_chunks.extend(additional_chunks)

        retrieved_chunks = remove_duplicates(retrieved_chunks)


        # retrieved_chunks ---> retrieved_chunks_texts

        if self.replace_retrieved_chunks_with_parent_paragraph:
            paragraphs = [paragraph for chunk in retrieved_chunks for paragraph in chunk.Paragraphs]

            paragraphs = remove_duplicates(paragraphs)
            retrieved_chunks_texts = [paragraph.Text for paragraph in paragraphs]

        else:
            retrieved_chunks_texts = [chunk.Text for chunk in retrieved_chunks]

        # if sum_of_retrieved_token_length_limit > 0:
        #     total_tokens = 0
        #     filtered_chunks = []
        #     for chunk in retrieved_chunks_texts:
        #         c_len = token_len(chunk)
        #         filtered_chunks.append(chunk)
        #         if total_tokens + c_len > sum_of_retrieved_token_length_limit:
        #             break
        #         total_tokens += c_len
        #     retrieved_chunks_texts = filtered_chunks


        return retrieved_chunks_texts

    def _ask_notTool(self, query, top_k=30, include_prompt=True):
        if self.answer_mode=='none':
            self.answer_mode = 'classic'
        elif self.answer_mode=='tool':
            print("changing mode to classic, chat history is cleared")
            self.llm.reset_history()

        # encode query -> retrieve top_k chunks -> generate answer

        retrieved_docs = self.retrieve(query, top_k)

        prompt,response = self.llm.rag_generate(query=query, retrieved_chunks=retrieved_docs)


        if include_prompt:
            answer = prompt + "\n\n# Answer:\n" + response
        else:
            answer = response

        return answer

    def ask(self, query, top_k=30, include_prompt=False,useTool=False):
        if self.llm is None:
            print("Error: LLM is not set. Please set the LLM before asking a question. or use retrieve()")
            return ""

        if useTool:
            return self._ask_tool(query)
        else:
            return self._ask_notTool(query, top_k=top_k, include_prompt=include_prompt)

    def _ask_tool(self, query):
        print("using tool")
        if self.answer_mode=='none':
            self.answer_mode = 'tool'
        elif self.answer_mode=='classic':
            print("changing mode to tool, chat history is cleared")
            self.llm.reset_history()

        response = self.llm.generate(prompt=query,tool_function= self.retrieve_interface)
        return response

    def clear_history(self):
        self.llm.reset_history()

    def retrieve_interface(self, query:str, top_k:int) -> str:
        """retrieves k chunks based on query.
        The chunks are then merged and returned as a single string.
        if the retrieved chunks didn't contain the information you needed, try to increase the top_k value or use different query.
        Args:
            query: The query to retrieve chunks for.
            top_k: The number of chunks to retrieve. recommended to be larger than 20.
        """
        retrieved_chunks=self.retrieve(query=query, top_k=top_k)

        retrieved_chunks_text=LLMIdentity.merge_chunks(retrieved_chunks=retrieved_chunks)

        retrieved_chunks_text += f"\n if the retrieved chunks didn't contain the information you needed, try to increase the top_k value or use different query. call the tool again right now with function_call before talking to the user."

        return retrieved_chunks_text






def cosine_similarity(a, b):
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))



from MetaRagTool.RAG.Chunkers import ChunkingMethod

class MetaRagConfig:
    def __init__(self, encoder=None, contexts=None, qas=None, llm=None, splitting_method=ChunkingMethod.SENTENCE_MERGER,
                 chunk_size=90, chunk_overlap=3, top_k=20, use_neighbor_embeddings=False
                 , add_neighbor_chunks=False, add_neighbor_chunks_smart=False,
                 breath_first_retrival=False
                 , depth_first_retrival=False, replace_retrieved_chunks_with_parent_paragraph=False
                 , normalize_text=False, use_parentParagraph_embeddings=False, max_sentence_len=-1,reranker=None,rerank=False,
                 additional_top_k=5, include_reflections=False,
                 include_reflections_k=1, include_refractions=False, sent_merge_merged_chunks=True, multi_hop=False,
                 log_chunking_report=False, weighted_bfs=False,
                 embedding_steering_influence_factor=0.35,
                 add_neighbor_chunks_k=2, judge=None, useTool=False, project_name=None, run_name=None,
                 encoder_name=None, sample_size=None, qa_sample_size=None,
                 local_mode=None,
                 multi_hop_hardness_factor=None, judged=None, llm_name=None,fine_grain_progressbar=True

                 ):
        self.fine_grain_progressbar = fine_grain_progressbar
        self.encoder = encoder
        self.contexts = contexts
        self.qas = qas
        self.llm = llm
        self.splitting_method = splitting_method
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.top_k = top_k
        self.use_neighbor_embeddings = use_neighbor_embeddings
        self.add_neighbor_chunks = add_neighbor_chunks
        self.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        self.breath_first_retrival = breath_first_retrival
        self.depth_first_retrival = depth_first_retrival
        self.replace_retrieved_chunks_with_parent_paragraph = replace_retrieved_chunks_with_parent_paragraph
        self.normalize_text = normalize_text
        self.use_parentParagraph_embeddings = use_parentParagraph_embeddings
        if max_sentence_len == -1:
            max_sentence_len = chunk_size
        self.max_sentence_len = max_sentence_len
        self.additional_top_k = additional_top_k
        self.include_reflections = include_reflections
        self.include_reflections_k = include_reflections_k
        self.include_refractions = include_refractions
        self.sent_merge_merged_chunks = sent_merge_merged_chunks
        self.multi_hop = multi_hop
        self.log_chunking_report = log_chunking_report
        self.weighted_bfs = weighted_bfs
        self.embedding_steering_influence_factor = embedding_steering_influence_factor
        self.add_neighbor_chunks_k = add_neighbor_chunks_k
        self.judge = judge
        self.useTool = useTool
        self.project_name = project_name
        self.run_name = run_name

        self.encoder_name = encoder_name
        self.sample_size = sample_size
        self.qa_sample_size = qa_sample_size
        self.local_mode = local_mode
        self.multi_hop_hardness_factor = multi_hop_hardness_factor
        self.judged = judged
        self.llm_name = llm_name
        self.reranker = reranker
        self.rerank = rerank

    def toDict(self):
        return {
            "splitting_method": self.splitting_method,
            "chunk_size": self.chunk_size,
            "chunk_overlap": self.chunk_overlap,
            "top_k": self.top_k,
            "use_neighbor_embeddings": self.use_neighbor_embeddings,
            "add_neighbor_chunks": self.add_neighbor_chunks,
            "add_neighbor_chunks_smart": self.add_neighbor_chunks_smart,
            "breath_first_retrival": self.breath_first_retrival,
            "depth_first_retrival": self.depth_first_retrival,
            "replace_retrieved_chunks_with_parent_paragraph": self.replace_retrieved_chunks_with_parent_paragraph,
            "normalize_text": self.normalize_text,
            "use_parentParagraph_embeddings": self.use_parentParagraph_embeddings,
            "max_sentence_len": self.max_sentence_len,
            "additional_top_k": self.additional_top_k,
            "include_reflections": self.include_reflections,
            "include_reflections_k": self.include_reflections_k,
            "include_refractions": self.include_refractions,
            "sent_merge_merged_chunks": self.sent_merge_merged_chunks,
            "multi_hop": self.multi_hop,
            "weighted_bfs": self.weighted_bfs,
            "embedding_steering_influence_factor": self.embedding_steering_influence_factor,
            "add_neighbor_chunks_k": self.add_neighbor_chunks_k,
            "rerank": self.rerank,
            "reranker_name": self.reranker.model_name if self.rerank else None,
            "useTool": self.useTool,
            "encoder_name": self.encoder_name,
            "sample_size": self.sample_size,
            "qa_sample_size": self.qa_sample_size,
            "local_mode": self.local_mode,
            "multi_hop_hardness_factor": self.multi_hop_hardness_factor,
            "judged": self.judged,
            "llm_name": self.llm_name,

        }


    def reset_rag_attributes(self):
        defaultConfig = MetaRagConfig()
        self.log_chunking_report = defaultConfig.log_chunking_report
        self.sent_merge_merged_chunks = defaultConfig.sent_merge_merged_chunks
        self.add_neighbor_chunks_k = defaultConfig.add_neighbor_chunks_k
        self.embedding_steering_influence_factor = defaultConfig.embedding_steering_influence_factor
        self.splitting_method = defaultConfig.splitting_method
        self.chunk_size = defaultConfig.chunk_size
        self.chunk_overlap = defaultConfig.chunk_overlap
        self.use_neighbor_embeddings = defaultConfig.use_neighbor_embeddings
        self.use_parentParagraph_embeddings = defaultConfig.use_parentParagraph_embeddings
        self.add_neighbor_chunks = defaultConfig.add_neighbor_chunks
        self.add_neighbor_chunks_smart = defaultConfig.add_neighbor_chunks_smart
        self.breath_first_retrival = defaultConfig.breath_first_retrival
        self.weighted_bfs = defaultConfig.weighted_bfs
        self.depth_first_retrival = defaultConfig.depth_first_retrival
        self.additional_top_k = defaultConfig.additional_top_k
        self.replace_retrieved_chunks_with_parent_paragraph = defaultConfig.replace_retrieved_chunks_with_parent_paragraph
        self.normalize_text = defaultConfig.normalize_text
        self.include_reflections = defaultConfig.include_reflections
        self.include_reflections_k = defaultConfig.include_reflections_k
        self.include_refractions = defaultConfig.include_refractions
        self.rerank = defaultConfig.rerank


def get_baseline_config():
    return MetaRagConfig(splitting_method=ChunkingMethod.RECURSIVE)


def get_best_config():
    return MetaRagConfig(splitting_method=ChunkingMethod.SENTENCE_MERGER,
                         use_neighbor_embeddings=True, use_parentParagraph_embeddings=True,
                         replace_retrieved_chunks_with_parent_paragraph=True,
                         add_neighbor_chunks_smart=True,
                         embedding_steering_influence_factor=0.2)



import weave
from MetaRagTool.LLM.LLMIdentity import LLMIdentity
from openai import OpenAI
import time

from MetaRagTool import Constants

class OpenaiGpt(LLMIdentity):


    endpoint_azure = "https://models.inference.ai.azure.com"
    endpoint_openrouter = "https://openrouter.ai/api/v1"
    MODEL_deepseek_v3 = "deepseek/deepseek-chat-v3-0324:free"
    MODEL_llama_4_maverick = "meta-llama/llama-4-maverick:free"
    MODEL_llama_3="meta-llama/llama-3.3-70b-instruct:free"
    MODEL_gpt_4o = "gpt-4o"

    def __init__(self, model_name=MODEL_deepseek_v3, has_memory=True, custom_system_message=None,
                 RequestPerMinute_limit=15, endpoint=endpoint_openrouter, api_key=None):
        if api_key is None:
            api_key = Constants.API_KEY_OPENROUTER
        super().__init__(model_name=model_name, has_memory=has_memory, custom_system_message=custom_system_message,
                         RequestPerMinute_limit=RequestPerMinute_limit)
        self.client = OpenAI(
            base_url=endpoint,
            api_key=api_key,
        )

        self.reset_history()

    def reset_history(self):
        system_instruction = self.GetCorrectSystemMessage(None)
        self.messages_history = [
            {"role": "system", "content": system_instruction},
            # {"role": "system","content": [{"type": "text","text": system_instruction}]}

        ]

    @weave.op()
    def generate(self, prompt: str, query_to_be_saved: str = None, tool_function=None) -> str:
        if tool_function is not None:
            print("OpenaiGpt does not support tool use")

        if self.RequestPerMinute_limit > 0 and tool_function is not None:
            self.manage_rpm()

        self.messages_history.append({"role": "user", "content": prompt})
        # self.messages_history.append( {"role": "user", "content": [{"type": "text", "text": prompt}]})



        while True:
            try:
                response = self.client.chat.completions.create(
                    messages=self.messages_history,
                    model=self.model_name,
                    # tools=
                )
                text_output=response.choices[0].message.content
                break
            except Exception as e:
                print(f"Error sending message, waiting 10 seconds to retry")
                time.sleep(10)




        if not self.has_memory:
            self.reset_history()

        self.messages_history.append(response.choices[0].message)

        return text_output



import MetaRagTool.Constants as Constants
import pandas as pd
from datasets import load_from_disk,load_dataset





def loadWikiFaQa(sample_size=-1, multi_hop = False,
                 random_seed=50, multi_hop_hardness_factor=0,custom_path=None,qa_sample_size=-1):

    if Constants.local_mode:
        path = 'D:\Library\MSc Ai\Thesis\Persian RAG\Datasets\synthetic data\\final\WikiFaQA'
        if multi_hop:
            path = f'D:\Library\MSc Ai\Thesis\Persian RAG\Datasets\synthetic data\\final\WikiFaQA_multiHop'

        existing_dataset = load_from_disk(path)
    else:
        path = 'codersan/WikiFaQA'
        if multi_hop:
            path = 'codersan/WikiFaQA_multiHop'
        if custom_path is not None:
            path = custom_path

        dataset_dict = load_dataset(path)
        existing_dataset = dataset_dict["train"]




    if sample_size != -1 and sample_size < len(existing_dataset):
        existing_dataset = existing_dataset.shuffle(seed=random_seed)
        existing_dataset = existing_dataset.select(range(sample_size))

    contexts = existing_dataset['context']
    exactPairs = existing_dataset['exactPairs']

    qa_data = []
    for pair_list in exactPairs:
        for pair in pair_list:
            if multi_hop:
                qa_data.append({
                    'question': pair['question'],
                    'answer1': pair['answer1'],
                    'answer2': pair['answer2'],
                    'distance': pair['distance'],
                })
            else:
                qa_data.append({
                    'question': pair['question'],
                    'answer': pair['answer']
                })

    qa = pd.DataFrame(qa_data)
    if multi_hop_hardness_factor==1:
        # top half of qa
        qa = qa[qa['distance'] >1077]


    elif multi_hop_hardness_factor==2:
        # top quarter of qa
        qa = qa[qa['distance'] > 4963]

    if qa_sample_size != -1 and qa_sample_size < len(qa):
        qa = qa.sample(qa_sample_size, random_state=random_seed)


    return contexts, qa





def loadFarSick(relatedness_score_threshold=4.5, sample_size=-1):
    file_path = 'D:\Library\MSc Ai\Thesis\Persian RAG\Datasets\FarSick.txt'
    data = pd.read_csv(file_path, sep='\t')
    data = data.drop_duplicates()
    data = data[data['relatedness_score'] >= relatedness_score_threshold]

    if sample_size != -1 and sample_size < len(data):
        data = data.sample(sample_size)

    allSentences = pd.concat([data['sentence_A'], data['sentence_B']], ignore_index=True)
    allSentences = allSentences.drop_duplicates()

    sentencePairs = data[['sentence_A', 'sentence_B']]

    print("dataset loaded successfully")

    return allSentences, sentencePairs


def loadPersianQa():
    file_path = 'D:\Library\MSc Ai\Thesis\Persian RAG\Datasets\persian_qa_cleaned.csv'
    data = pd.read_csv(file_path)
    data = data.drop_duplicates()
    data = data[['question', 'context', 'answers_raw']]

    print("dataset loaded successfully")

    return data
